import os, sys
from collections import Counter
from typing import Set, Generator, Dict, Any
from owlapy.class_expression import *
from ontolearn.owl_neural_reasoner import TripleStoreNeuralReasoner
from owlapy.owl_property import (
    OWLDataProperty,
    OWLObjectInverseOf,
    OWLObjectProperty,
    OWLProperty,
)
from owlapy.iri import IRI
from owlapy import owl_expression_to_dl
from owlapy.parser import DLSyntaxParser
from ontolearn.knowledge_base import KnowledgeBase
from graphviz import Digraph
import json
import shutil
import random


class ReasoningNode:
    """ A node representing a step in the reasoning tree """
    def __init__(self, expression: Any, instances: Set[Any] = None):
        self.expression = expression
        self.instances = instances if instances is not None else set()
        self.children = []

    def add_child(self, child):
        self.children.append(child)

    def to_dict(self):
        """ Convert tree to a dictionary format for easy visualization """
        return {
            "expression": str(self.expression),
            "instances": [str(inst) for inst in self.instances],
            "children": [child.to_dict() for child in self.children]
        }


def build_reasoning_tree(reasoner, expression: Any) -> ReasoningNode:
    """ Builds a tree representing the reasoning process without modifying the reasoner """
    root = ReasoningNode(expression)
    _populate_tree(reasoner, root, expression)
    return root


def _populate_tree(reasoner, node: ReasoningNode, expression: Any):
    """ Recursively populates the reasoning tree using the existing reasoner """
    if isinstance(expression, (OWLClass, OWLObjectOneOf)):
        node.instances = set(reasoner.instances(expression))
    elif isinstance(expression, OWLObjectUnionOf):
        for op in expression.operands():
            child_node = ReasoningNode(op)
            node.add_child(child_node)
            _populate_tree(reasoner, child_node, op)
        node.instances = set.union(*(child.instances for child in node.children))
    elif isinstance(expression, OWLObjectIntersectionOf):
        for op in expression.operands():
            child_node = ReasoningNode(op)
            node.add_child(child_node)
            _populate_tree(reasoner, child_node, op)
        node.instances = set.intersection(*(child.instances for child in node.children))
    elif isinstance(expression, OWLObjectComplementOf):
        child_node = ReasoningNode(expression.get_operand())
        node.add_child(child_node)
        _populate_tree(reasoner, child_node, expression.get_operand())
        all_individuals = {i for i in reasoner.individuals_in_signature()}
        node.instances = all_individuals - child_node.instances
    elif isinstance(expression, OWLObjectSomeValuesFrom):
        object_property = expression.get_property()
        filler_expression = expression.get_filler()
        filler_node = ReasoningNode(filler_expression)
        node.add_child(filler_node)
        _populate_tree(reasoner, filler_node, filler_expression)

        result = Counter()
        for object_individual in filler_node.instances:
            subjects = reasoner.get_individuals_with_object_property(
                obj=object_individual,
                object_property=object_property
            )
            result.update(subjects)

        node.instances = {ind for ind, count in result.items() if count >= 1}
    elif isinstance(expression, OWLObjectAllValuesFrom):
        object_property = expression.get_property()
        filler_expression = expression.get_filler()
        complement_node = ReasoningNode(OWLObjectComplementOf(filler_expression))
        node.add_child(complement_node)
        _populate_tree(reasoner, complement_node, OWLObjectComplementOf(filler_expression))

        negation_some = OWLObjectSomeValuesFrom(object_property, OWLObjectComplementOf(filler_expression))
        negation_some_node = ReasoningNode(negation_some)
        node.add_child(negation_some_node)
        _populate_tree(reasoner, negation_some_node, negation_some)

        all_individuals = {i for i in reasoner.individuals_in_signature()}
        node.instances = all_individuals - negation_some_node.instances
    elif isinstance(expression, OWLObjectMaxCardinality):
        object_property = expression.get_property()
        filler_expression = expression.get_filler()
        cardinality = expression.get_cardinality()

        filler_node = ReasoningNode(filler_expression)
        node.add_child(filler_node)
        _populate_tree(reasoner, filler_node, filler_expression)

        subject_counts = {ind: 0 for ind in reasoner.individuals_in_signature()}
        for object_ind in filler_node.instances:
            subjects = reasoner.get_individuals_with_object_property(obj=object_ind, object_property=object_property)
            for subj in subjects:
                subject_counts[subj] += 1

        node.instances = {ind for ind, count in subject_counts.items() if count <= cardinality}
    else:
        raise NotImplementedError(f"Instances for {type(expression)} are not implemented yet")
    


def format_instances(instances, n=3):
    """Format instances: Show at most 4, followed by '...' if more exist."""
    instance_names = [inst.str.split('#')[-1] for inst in instances]
    if len(instance_names) > n:
        instance_names = instance_names[:n]
        instance_names.append('...')
        return "{" + ", ".join(instance_names) + "}"
    return str(set(instance_names))


def visualize_reasoning_tree_graphviz(root: ReasoningNode, filename="reasoning_tree"):
    """ Uses Graphviz to visualize the reasoning tree with the concept at the top 
        and its retrieved instances displayed separately. """
    dot = Digraph(format="png")  # Create a directed graph

    root_id = str(id(root))
    
    # Root node: Show the full concept at the top in red
    concept_label = f"{owl_expression_to_dl(root.expression)}"
    dot.node(root_id, concept_label, shape="box", style="bold", color='red')

    # Instances node: Show retrieved instances in a separate node
    instance_label = format_instances(root.instances)
    instance_node_id = f"{root_id}_instances"
    dot.node(instance_node_id, instance_label, shape="ellipse", style="bold", color="red")

    # Add a horizontal edge from concept node to instances
    dot.edge(root_id, instance_node_id, label="output", dir="forward", arrowhead="normal", constraint="false", style="dashed")

    def _add_nodes(node, parent=None, edge_label=""):
        """ Recursively add nodes to the Graphviz object """
        node_id = str(id(node))

        if node is not root:  # Skip re-adding the root
            if not node.children:  
                # Leaf nodes: If it's a final retrieved instance, highlight it in red
                node_label = format_instances(node.instances)
                dot.node(node_id, node_label, shape="ellipse", style="bold", color='blue')
            else:
                # Normal child nodes: Show only instances
                node_label = format_instances(node.instances)
                dot.node(node_id, node_label, shape="ellipse")

            if parent:
                dot.edge(str(id(parent)), node_id, label=edge_label)  # Label edges with the concept breakdown

        for child in node.children:
            _add_nodes(child, node, edge_label=owl_expression_to_dl(child.expression))

    _add_nodes(root)
    dot.render(filename, view=False)



def visualize_reasoning_tree(root: ReasoningNode):
    """ Print reasoning tree in a structured way with at most n instances displayed per node. """
    def _print_tree(node, level=0):
        #out_tree = ""
        instances = format_instances(node.instances)
        #print(" " * (level * 4) + f"- {owl_expression_to_dl(node.expression)} -----> {instances}")
        out_tree = " " * (level * 4) + f"- {owl_expression_to_dl(node.expression)} -----> {instances}"
        for child in node.children:
            _print_tree(child, level + 1)
        return out_tree

    return _print_tree(root)


def generate_trees(root_dir: str, exprs_path: str, path_embs: str, out_dir: str):
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
        os.makedirs(out_dir+"/img")
    def is_valid_expr(key):
        if "¬" not in key and "∃" not in key and "∀" not in key and '⁻' not in key and '≤' not in key and '≥' not in key and "{" not in key and len(key.split(" ")) == 1:
            return False
        return True
    
    kb_path = root_dir+"/kb/ontology.owl"
    reasoner = TripleStoreNeuralReasoner(path_of_kb=kb_path, path_neural_embedding=path_embs, gamma=0.9)
    kb = KnowledgeBase(path=kb_path)
    shutil.copy(kb_path, f"{out_dir}/ontology.owl")
    kb_namespace = list(kb.ontology.classes_in_signature())[0].str
    if "#" in kb_namespace:
        kb_namespace = kb_namespace.split("#")[0] + "#"
    elif "/" in kb_namespace:
        kb_namespace = kb_namespace[:kb_namespace.rfind("/")] + "/"
    elif ":" in kb_namespace:
        kb_namespace = kb_namespace[:kb_namespace.rfind(":")] + ":"
    dls_parser = DLSyntaxParser(kb_namespace)
    with open(exprs_path) as f:
        class_exprs = json.load(f)
    out_data = {}
    if isinstance(class_exprs[0], list):
        class_exprs = dict(class_exprs)
    valid_exprs = [key for key in class_exprs if is_valid_expr(key)]
    random.shuffle(valid_exprs)
    class_exprs = {key: class_exprs[key] for key in valid_exprs}
    i = 0
    for expr in class_exprs:
        out_data[expr] = class_exprs[expr]
        concept = dls_parser.parse(expr)
        try:
            reasoning_tree = build_reasoning_tree(reasoner, concept)
            tree_as_str = visualize_reasoning_tree(reasoning_tree) 
            visualize_reasoning_tree_graphviz(reasoning_tree, f"{out_dir}/img/img_{i}")
            with open(f"{out_dir}/img/img_{i}") as f_tree:
                out_data[expr]["tree_as_str"] = f_tree.read()
            out_data[expr]["tree_img_path"] = f"{out_dir}/img/img_{i}"
        except:
            continue
        i += 1
        if i >= 500:
            break
        

    for file_name in os.listdir(f"{out_dir}/img/"):
        if not file_name.endswith(".png"):
            os.remove(f"{out_dir}/img/{file_name}")
    with open(f"{out_dir}/class_expressions.json", "w") as f:
        json.dump(out_data, f)

if __name__=="__main__":
    random.seed(42)
    if len(sys.argv) < 4:
        raise ValueError("You must provide the ```knowledge base path```, ```output directory``` and ```the path to embeddings``` in order")
    root_dir = sys.argv[1]
    out_dir = sys.argv[2]
    path_embs = sys.argv[3]
    exprs_path = f"{root_dir}"+"/training_data/training_data.json"
    generate_trees(root_dir, exprs_path, path_embs, out_dir)

"""
EBR = TripleStoreNeuralReasoner(path_of_kb="KGs/Family/father.owl", gamma=0.9)
reasoner = EBR  
A = OWLObjectAllValuesFrom(property=OWLObjectProperty(IRI('http://example.com/father#','hasChild')),filler=OWLObjectComplementOf(OWLClass(IRI('http://example.com/father#','male'))))

# A= OWLClass(IRI('http://example.com/father#','person'))
B = OWLObjectComplementOf(OWLClass(IRI('http://example.com/father#','person')))
concept = OWLObjectIntersectionOf([A, A])  # Example OWL expression
reasoning_tree = build_reasoning_tree(reasoner, concept)
visualize_reasoning_tree(reasoning_tree) 
visualize_reasoning_tree_graphviz(reasoning_tree, "reasoning_tree_output")
"""